/*
 *      ISIS release V1.1, Dec. 1988
 *      Export restrictions apply
 */
#define   ISIS_SYS
# include "isis.h"

#ifndef  SIMWRITEV

#       define MAXIOVECLEN     16

#else   SIMWRITEV

#       define MAXIOVECLEN     1

static writev(sock, iovp, iovl)
  register sock;
  register iovec *iovp;
  {
        if(iovl != 1)
            panic("writev iovl != 1!\n");
        return(write(sock, iovp->iov_base, iovp->iov_len);
  }
#endif  SIMWRITEV

extern  errno;

/* Algorithm for writing a message to a FILE */
msg_fwrite(file, mp)
  FILE *file;
  register message *mp;
  {
        register nb, len = 0, iovlen;
        register struct iovec *iovp;
        register iov_len;
        register char *iov_base;
        t_scheck();
        iovlen = msg_getiovlen(mp);
        iovp = msg_getiovec(mp);
        iov_len = iovp->iov_len;
        iov_base = iovp->iov_base;
        while(iovlen)
        {
            while((nb = fwrite(iov_base, 1, iov_len, file)) == -1)
            {
                if(errno == EINTR)
                    continue;
                return(-1);
            }
            len += nb;
            while(nb)
            {
                if(iov_len <= nb)
                {
                    nb -= iov_len;
                    --iovlen;
                    ++iovp;
                    iov_len = iovp->iov_len;
                    iov_base = iovp->iov_base;
                }
                else
                {
                    iov_base += nb;
                    iov_len -= nb;
                    nb = 0;
                }
            }
        }
        if(len != msg_getlen(mp))
            panic("algorithmic error in msg_fwrite");
        return(0);
  }

message *
msg_fread(file)
  FILE *file;
  {
        register message *mp;
        register long *space;
        register char *ptr;
        int length = 0;
        register nb;
        block_desc *blk_desc;

        while((nb = fread((char*)&length, 1, sizeof(long), file)) != sizeof(long))
        {
            if(nb == -1 && errno == EINTR)
                continue;
            return(0);
        }
        length = ntohl (length);
        blk_desc = msg_blockalloc(length, 0);
        space = (long *)blk_desc->addr;
        space[0] = htonl (length);
        ptr = (char*)&space[1];
        length -= sizeof(long);
        while(length)
        {
            while((nb = fread(ptr, 1, length, file)) <= 0)
            {
                if(nb == -1 && errno == EINTR)
                    continue;
                return(0);
            }
            length -= nb;
            ptr += nb;
        }
        if(mp = msg_reconstruct((char*)space, blk_desc))
            return(mp);
        return(0);
  }

msg_write(sock, mp)
  register message *mp;
  {
        register nb, len = 0, iovlen;
        register struct iovec *iovp, *iovb;
        t_scheck();
        iovlen = msg_getiovlen(mp);
        iovb = iovp = (struct iovec*)malloc(nb = sizeof(struct iovec)*iovlen);
        bcopy(msg_getiovec(mp), iovp, nb);
        while(iovlen)
        {
            register iovl = iovlen;
            if(iovl > MAXIOVECLEN)
                iovl = MAXIOVECLEN;
            while((nb = writev(sock, iovp, iovl)) == -1)
            {
                if(errno == EINTR)
                    continue;
                if(sock == isis_socket)
                {
                    perror("isis_write");
                    isis_has_crashed("isis_write");
                }
                free(iovb);
                return(-1);
            }
            len += nb;
            while(nb)
            {
                if(iovp->iov_len <= nb)
                {
                    nb -= iovp->iov_len;
                    --iovlen;
                    ++iovp;
                }
                else
                {
                    iovp->iov_base += nb;
                    iovp->iov_len -= nb;
                    nb = 0;
                }
            }
        }
        if(len != msg_getlen(mp))
            panic("algorithmic error in msg_write");
        free(iovb);
        return(0);
  }

message *
msg_read(sock)
  {
        register message *mp;
        register long *space;
        register char *ptr;
        int length = 0;
        register nb;
        block_desc *blk_desc;
        static ntries;

        while((nb = read(sock, (char*)&length, sizeof(long))) != sizeof(long))
        {
            if(nb == -1 && errno == EINTR)
                continue;
            if(sock == isis_socket && (nb < 0 || ntries++ == 10))
                isis_has_crashed("read failed");
            return(0);
        }
        ntries = 0;
        length = ntohl (length);
        blk_desc = msg_blockalloc(length, 0);
        space = (long *)blk_desc->addr;
        space[0] = htonl (length);
        ptr = (char*)&space[1];
        length -= sizeof(long);
        while(length)
        {
            while((nb = read(sock, ptr, length)) <= 0)
            {
                if(nb == -1 && errno == EINTR)
                    continue;
                if(sock == isis_socket)
                {
                    if(nb == 0 && ntries++ < 10) 
                    {    
                        sleep(1);
                        continue; 
                    }
                    isis_has_crashed("read after fragmentation failed");
                }
                return(0);
            }
            length -= nb;
            ptr += nb;
        }
        if(mp = msg_reconstruct((char*)space, blk_desc))
            return(mp);
        if(sock == isis_socket)
            isis_has_crashed("can't reconstruct");
        return(0);
  }
